import numpy as np
from PIL import Image, ImageEnhance
import torchvision
import torch
import zxingcpp
from barcode_reader import unet, adder
import importlib.resources
from autocapture_model import detector


class AdderDecoder:
    def __init__(self, model_path=None, num_nets=5, magnitude=16, cuda=False, cropped=False):
        self.untransform = torchvision.transforms.ToPILImage()
        self.image_size = (256, 1024)
        if cuda:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = torch.device('cpu')
        if model_path is not None:
            add = adder.AdderDiffusion(num_nets=num_nets, model_path=model_path, pretrained=False, magnitude=magnitude)
        else:
            path = importlib.resources.open_binary('barcode_reader.models', 'adder.pytorch.model')
            add = adder.AdderDiffusion(num_nets=num_nets, model_path=path, magnitude=magnitude, cuda=cuda)
        self.model = add.model
        self.model.eval().to(self.device)
        self.cropped = cropped
        if not self.cropped:
            self.document_detector = detector.DocumentDetector()

    def extract_barcode(self, im, box=None):
        if type(im) is torch.Tensor:
            im = torchvision.transforms.ToPILImage()(im)
        im_arr = np.array(im)
        if self.cropped:
            return im
        if box is None:
            result = self.document_detector.detect_document(name='test', image=im)
            if result[0] is True:
                if 'barcode' in result[1]['raw_result'].keys():
                    box = result[1]['raw_result']['barcode']['original_box']
                else:
                    return None
        padding = int((0.15) * (box[3] - box[1]))
        pad_x, pad_y = padding, padding
        return im_arr[box[1]-pad_y:box[3]+pad_y, box[0]-pad_x:box[2]+pad_x]

    def _input_transform(self, im):
        pad_top, pad_left, pad_bottom, pad_right = 10, 15, 10, 15
        pad_top, pad_left, pad_bottom, pad_right = 0, 0, 0, 0
        transform = torchvision.transforms.Compose([
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Resize([self.image_size[0]-(pad_top + pad_bottom), self.image_size[1]-(pad_left + pad_right)]),
            torchvision.transforms.Pad(padding=[pad_left, pad_top, pad_right, pad_bottom], padding_mode='constant', fill=1)
        ])
        return transform(im)

    def decode(self, im, box=None):
        im = self.extract_barcode(im, box)
        if im is None:
            return None
        im = self._input_transform(im)
        self.input_image = im
        im = im.unsqueeze(0).to(self.device)
        with torch.no_grad():
            out = self.model(im)
        out = out[0]
        out = torch.clamp(out, min=0, max=1)
        out = self.untransform(out)
        self.output_image = out
        result = zxingcpp.read_barcode(out)
        if result is not None:
            return result.text
        return None



import torch
import torchvision
import sys
import os
import random
import zxingcpp
import datetime
import time
from barcode_reader.unet import UNet, L1_DICE, VGGPerceptualLoss
import argparse


class DiffDataset(torch.utils.data.Dataset):
    def __init__(self, root='/data/doc_detect/barcodes/barcode_diffusion/',
                 augment_flag=True,
                 limit=None,
                 mode='train'):
        self.targets, self.features, self.filenames = [], [], []
        self.augment_flag = augment_flag
        self.mode = mode
        self.root = root
        self.image_size = [256, 1024]
        self._load_dataset(mode=mode, limit=limit)

    def _get_files(self, mode):
        files = []
        if mode is None:
            files = os.listdir(self.root + 'features/')
        elif mode == 'train':
            test = self._get_files(mode='test')
            for file in os.listdir(self.root + 'features/'):
                if file not in test:
                    files.append(file)
            del test
        else:
            with open(self.root + mode, 'r') as f:
                lines = f.readlines()
                for line in lines:
                    files.append(line[:line.find("'\'")])
        return files

    def _load_dataset(self, mode, limit):
        files = self._get_files(mode=mode)
        for file in os.listdir(self.root + 'features/'):
            if file in files:
                self.filenames.append(file)
            if limit is not None and len(self.filenames) >= limit:
                break

    def _load_one(self, file):
        target = torch.load(self.root + 'targets/' + file)
        feature = torch.load(self.root + 'features/' + file)
        return feature, target

    def _augment(self, x, y):
        x, y = torchvision.transforms.Resize(self.image_size)(x), torchvision.transforms.Resize(self.image_size)(y)
        pad_y_amount, pad_x_amount = random.randint(0, 30), random.randint(0, 60)
        pad_top, pad_left = random.randint(0, pad_y_amount), random.randint(0, pad_x_amount)
        pad_bottom, pad_right = pad_y_amount - pad_top, pad_x_amount - pad_left
        if self.augment_flag:
            t = torchvision.transforms.Compose([
                torchvision.transforms.Resize([self.image_size[0]-(pad_top + pad_bottom),
                                               self.image_size[1]-(pad_left + pad_right)]),
                torchvision.transforms.Pad(padding=[pad_left, pad_top, pad_right, pad_bottom], padding_mode='constant', fill=1),
            ])
        else:
            t = torchvision.transforms.Compose([
                torchvision.transforms.Resize(self.image_size)
            ])
        return t(x), t(y)

    def __getitem__(self, index):
        feature, target = self._load_one(self.filenames[index])
        feature, target = self._augment(feature, target)
        return feature, target, self.filenames[index], index

    def __len__(self):
        return len(self.filenames)


class UNetDecoder():
    def __init__(self, model_obj, device, num_nets=5):
        self.model = model_obj
        self.device = device
        self.model.eval().to(self.device)
        self.untransform = torchvision.transforms.ToPILImage()
        self.num_nets = num_nets

    def forward(self, x):
        return self.model(x)

    def decode_count(self, loader, lf=None, zxing_first=False, resize=False, limit=None):
        success, total = 0, 0
        rotation = []
        ave_loss, iterations = 0, 0
        if resize:
            untransform = torchvision.transforms.Compose([
                torchvision.transforms.Resize([256, 1024]),
                torchvision.transforms.ToPILImage()
            ])
        else:
            untransform = torchvision.transforms.ToPILImage()
        count = 0
        for i, (image_batch, target_batch, filename_batch, index_batch) in enumerate(loader):
            batch_size = len(filename_batch)
            count += batch_size
            if limit is not None:
                if count > limit:
                    break
            image_batch = torch.stack(image_batch).to(self.device)
            rotated = torchvision.transforms.functional.rotate(image_batch, 180)
            with torch.no_grad():
                out1 = self.forward(image_batch)
                if lf is not None:
                    target_batch = torch.stack(target_batch).to(self.device)
                    loss = torch.mean(torch.stack(lf(out1, target_batch)))
                    iterations += 1
                    ave_loss += loss.item()
            with torch.no_grad():
                out2 = self.forward(rotated)
            for i in range(len(out1)):
                norm1 = out1[i]
                norm2 = out2[i]
                norm1[norm1 > 1] = 1
                norm1[norm1 < 0] = 0
                norm2[norm2 > 1] = 1
                norm2[norm2 < 0] = 0
                norm1 = untransform(norm1)
                norm2 = untransform(norm2)
                try:
                    result = zxingcpp.read_barcode(norm1)
                    rotated_result = zxingcpp.read_barcode(norm2)
                except Exception:
                    result = None
                    rotated_result = None
                # norm1.save('temp.jpg')
                # result = self.zx.decode('temp.jpg', try_harder=True)
                # norm2.save('temp.jpg')
                # rotated_result = self.zx.decode('temp.jpg', try_harder=True)
                if result is not None:
                    success += 1
                elif rotated_result is not None:
                    rotation.append(filename_batch[i])
                total += 1
        if iterations > 0 and lf is not None:
            return success, total, success/total, rotation, ave_loss/iterations
        return success, total, success/total, rotation


class AdderDiffusion(torch.nn.Module):
    def __init__(self, num_nets=5, magnitude=16, model_path=None, mode='train', cuda=False):
        super().__init__()
        if cuda:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = torch.device('cpu')
        self.mode = mode
        self.num_nets = num_nets
        self.regularization_lambda = 1e-6
        layers = []
        for _ in range(num_nets):
            layers.append(UNet(in_channels=3, n_class=3, magnitude=magnitude, dropout_rate=0.0))
        self.model = torch.nn.Sequential(*layers)
        del layers
        if model_path is not None:
            self.load_model(model_path)

    def forward(self, x):
        return self.model(x)

    def collate_fn(self, batch):
        return tuple(zip(*batch))

    def save_model(self, path):
        torch.save(self.model.state_dict(), path)

    def load_model(self, path):
        sd = torch.load(path, map_location='cpu')
        self.model.load_state_dict(sd)
        print('model', path, 'loaded')

    def l2(self, loss):
        regularization_loss = torch.tensor(0.0).to(self.device)
        for net in self.model:
            for name, block in net.named_children():
                if 'encoder' in name:
                    regularization_loss += torch.norm(block.conv1.weight)
                    regularization_loss += torch.norm(block.conv2.weight)
                elif 'bridge' in name:
                    regularization_loss += torch.norm(block.weight)
                elif 'decoder' in name:
                    regularization_loss += torch.norm(block.conv1.weight)
                    regularization_loss += torch.norm(block.conv2.weight)
                    try:
                        regularization_loss += torch.norm(block.convT1.weight)
                    except Exception:
                        regularization_loss += torch.norm(block.conv3.weight)
        regularization_loss *= self.regularization_lambda
        loss += regularization_loss
        return loss, regularization_loss

    def train_one_epoch(self, epoch):
        header = f"Epoch: [{epoch}]"
        sum_loss, sum_l1, sum_dice, sum_vgg = 0, 0, 0, 0
        test_reg, sum_reg = 0, 0
        num_iterations = 0
        S = time.time()
        test_iterations, test_loss, test_l1, test_dice, test_vgg = 1, 0, 0, 0, 0
        self.model.eval()
        for ind, (image_batch, target_batch, _, index_batch) in enumerate(self.test_loader):
            image_batch = torch.stack(image_batch).to(self.device)
            target_batch = torch.stack(target_batch).to(self.device)
            with torch.no_grad():
                result = self.forward(image_batch)
                loss, dice, l1 = self.criterion(result, target_batch)
            vgg_loss = torch.tensor(0.0)
            reg_loss = torch.tensor(0.0)
            test_reg += reg_loss.item()
            test_iterations += 1
            test_loss += loss.item()
            test_l1 += l1.item()
            test_dice += dice.item()
            test_vgg += vgg_loss.item()
        self.model.train()
        for ind, (image_batch, target_batch, _, index_batch) in enumerate(self.train_loader):
            image_batch = torch.stack(image_batch).to(self.device)
            target_batch = torch.stack(target_batch).to(self.device)
            self.optimizer.zero_grad()
            result = self.forward(image_batch)
            loss, dice, l1 = self.criterion(result, target_batch)
            # vgg_loss, _, _ = self.vgg_loss(result, target_batch)
            vgg_loss = torch.tensor(0.0)
            # l1_cst = 1
            # loss = (vgg_cst) * vgg_loss + (l1_cst) * loss
            # loss, reg_loss = self.l2(loss)
            reg_loss = torch.tensor(0.0)
            sum_reg += reg_loss.item()
            num_iterations += 1
            sum_loss += loss.item()
            sum_l1 += l1.item()
            sum_dice += dice.item()
            sum_vgg += vgg_loss.item()
            loss.backward()
            # torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.0)
            self.optimizer.step()
        print(header, '[' + str(self.optimizer.param_groups[0]['lr']) + ']',
              'loss/l1/vgg/reg {:.4f}/{:.4f}/{:.4f}/{:.4f}'.format(sum_loss/num_iterations,
                                                                   sum_l1/num_iterations,
                                                                   sum_vgg/num_iterations,
                                                                   sum_reg/num_iterations),
              '----------------------------- val loss/l1/vgg/reg {:.4f}/{:.4f}/{:.4f}/{:.4f}'.format(test_loss/test_iterations,
                                                                                                     test_l1/test_iterations,
                                                                                                     test_vgg/test_iterations,
                                                                                                     test_reg/test_iterations),
              str(time.time()-S)[:7])
        return sum_loss/num_iterations, test_loss/test_iterations

    def _init_dataloaders(self, batch_size, train, test):
        self.train_loader = torch.utils.data.DataLoader(train, batch_size=batch_size, shuffle=True,
                                                        collate_fn=self.collate_fn, pin_memory=True,
                                                        prefetch_factor=3, num_workers=4)
        self.test_loader = torch.utils.data.DataLoader(test, batch_size=batch_size, shuffle=True,
                                                       collate_fn=self.collate_fn, pin_memory=True,
                                                       prefetch_factor=3, num_workers=4)

    def train(self, train, test, num_epochs=5000, batch_size=13, learning_rate=0.001, betas=(0.7, 0.99), version=18):
        print('Training', num_epochs, 'Epochs', 'Version:', version, 'Dataset:', len(train))
        self._init_dataloaders(batch_size=batch_size, train=train, test=test)
        params = [p for p in self.model.parameters()]
        self.optimizer = torch.optim.Adam(params, lr=learning_rate, betas=betas, eps=1e-08, weight_decay=0.0)
        self.criterion = L1_DICE(alpha=1.0)
        self.vgg_loss = VGGPerceptualLoss()
        for epoch in range(num_epochs):
            self.model.train().to(self.device)
            train_loss, test_loss = self.train_one_epoch(epoch)
            if epoch % 5 == 0:
                self.model.eval().to(self.device)
                self.log(version, train_loss, test_loss, epoch)

    def log(self, version, train_loss, test_loss, epoch, sample_size=None):
        s = time.time()
        self.model.eval().to(self.device)
        decoder = UNetDecoder(self.model, self.device, num_nets=self.num_nets)
        count, total, acc, rotation_candidates = decoder.decode_count(self.train_loader, lf=None)
        test = decoder.decode_count(self.test_loader, lf=None, limit=sample_size)
        model_name = datetime.datetime.today().strftime('%m%d-%H:%M')
        rotation_candidates += test[3]
        self.save_model('/home/nathan/barcode-reader/barcode_reader/models/adder/' + str(version) + '/' + model_name)
        print('     Model File:                            ', model_name)
        print('     Decoded train accuracy                  [' + str(count) + ', ' + str(total) + '] -', str(acc*100)[:5] + '%')
        print('     Decoded test accuracy                   [' + str(test[0]) + ', ' + str(test[1]) + '] -', str(test[2]*100)[:5] + '%')
        print('     Rotation candidates:                    [' + str(len(rotation_candidates)) + ']')
        print('Completed in:', time.time()-s)
        out_str = 'Epoch [' + str(epoch) + '] -'
        out_str += model_name + '- Train: ' + str(acc)[:7]
        out_str += ' - Test: ' + str(test[2])[:7] + ' - Loss: [' + str(train_loss)[:7] + ', ' + str(test_loss)[:7] + ']'
        out_str += ' Rotation Candidates: ' + str(len(rotation_candidates)) + '\n'
        with open('/home/nathan/barcode-reader/barcode_reader/models/adder/' + str(version) + '/model.log', 'a') as f:
            f.write(out_str)
        if epoch > 200 and acc > 0.5:
            with open('/home/nathan/barcode-reader/barcode_reader/models/adder/' + str(version) + '/rotation.log', 'a') as f:
                for filename in rotation_candidates:
                    f.write(filename + '\n')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--batch_size', default=13, type=int)
    parser.add_argument('--num_epochs', default=500, type=int)
    parser.add_argument('--limit', default=None, type=int)
    parser.add_argument('--version', default=35, type=int)
    parser.add_argument('--lr', default=0.0001, type=float)
    parser.add_argument('--num_nets', default=5, type=int)
    parser.add_argument('--size', default=512, type=int)
    parser.add_argument('--magnitude', default=16, type=int)
    args = parser.parse_args()
    S = time.time()
    print('Mag', args.magnitude, 'Num nets', args.num_nets)
    if args.limit is None:
        train_lim, test_lim = None, None
    else:
        train_lim, test_lim = int(0.9 * args.limit), int(0.1 * args.limit)
        if train_lim == 0:
            train_lim = 1
        if test_lim == 0:
            test_lim = 1
    train = DiffDataset(limit=train_lim,
                        mode='train',
                        augment_flag=True)
    test = DiffDataset(limit=test_lim,
                       mode='test',
                       augment_flag=False)
    print('Dataset loaded:', len(train), len(test), time.time()-S)
    diffusion = AdderDiffusion(num_nets=args.num_nets,
                               magnitude=args.magnitude,
                               cuda=True,
                               model_path='/home/nathan/barcode-reader/barcode_reader/models/adder/35/1120-12:44'
                               )
    diffusion.train(train=train,
                    test=test,
                    learning_rate=args.lr,
                    batch_size=args.batch_size,
                    num_epochs=args.num_epochs,
                    version=args.version,
                    )


# model 3 -> single UNet, 500 epochs
# model 4 -> double UNet, 170 - epochs, 76% train, 70% test, ~20% raw dataset
# model 5 -> quad UNet, up to 85% train, 80% test, then overfit down to 55% test
# model 6 -> quad UNet, stopped early, around 76% each
# model 7 -> quad UNet, stopped early (only 60 epochs), 65% each
# model 8 -> 5 Unets, multi gpu enabled, more than doubled dataset (L1 Loss)
    # maxed out around 77/67 train/test, collapsed after around 100 epochs
# model 9 -> 5 Unets, try out L2 loss, added dropout, adjusted betas
    # for model 1114-19:50
    # starts to drop off around 250 epochs at 80/75 train/test
    # gets close to 22% on raw dataset (+5%)
    # 1150 new from more_barcoces, 885 new from barcodes/train/features - 2035 total ()
# model 10 -> extension of model 9, changed back to l1 and removed negative padding, added ~2000 more images
    # ~ 700 new barcodes after 60 epochs of training (on top of model 9)
    # added slightly more dropout
# model 11 -> lowered padding amount, working on custom loss - l1_ssim
    # extended from model 10 1115-16:47
    # changed my mind on the padding, reset
    # adjusted alpha in loss function
# model 12 -> fixing stuff I broke
    # change data augmentation to happen within __getitem__ so that every epoch is a different variation
    # broke something in the process - i think it was my model definition using custom modules
    # ~3k new datapoints added
    # using ssim_l1 with low alpha (0.2) again since previous test was invalid
    # total data 11k
# model 13 -> testing out independent graph again
    # second changed alpha
    # third switched to l2
    # fourth turned up beta1 and learning rate, back to l1ssim
    # 5, l1 loss with sgd
# model 14 -> not started yet, just planning
    # once most of rotation is removed I'll restart training, use SGD and higher learning rate with decay
    # probably try just l1 loss again
    # using l1_dice

# baseline zxing test on full train dataset:
    # 3759 11146 0.33725103176027277
# baseline zxing test on docv2 barcodes test dataset:
    # 27.2%
# baseline zxing test on more_barcodes test dataset:
    # 48 %

    # model 19 ->
    # test 75%
    # docv2 36%

# model 35
# test 65 %


from urllib.parse import urljoin
import requests
import json
import datetime
import re
import io
from autocapture_model import detector
import numpy as np
import PIL.Image
import os

USERNAME = 'nathan'
PASSWORD =  'Driving a Volkswagon Vanagon is like stirring a pot~'
DAYS_BACK_RANGE = 1

today = datetime.datetime.today()
end_date = today - datetime.timedelta(days=DAYS_BACK_RANGE)

auth_body = {'requested_url':'', 'username': USERNAME,'passphrase': PASSWORD}

session = requests.Session()
response = session.post("https://mlweb-staging.ml.prod.trulioo.com/authenticate", data=auth_body)

request_body = {"end_date":end_date.isoformat(), "start_date":today.isoformat(), "target":"barcode"}
# request_body = {"target":"barcode"}
request_headers = {'Content-Type': 'application/json'}
# print(request_body)
response = requests.post("https://mlweb-staging.ml.prod.trulioo.com/get_results", data=json.dumps(request_body), headers=request_headers)

extract_id = re.findall(r'"id"[:] "[0-9a-zA-Z-]+', str(response.content))
all_images_id_list = [x[7:] for x in extract_id]

base_url = 'https://mlweb-staging.ml.prod.trulioo.com'

assert len(all_images_id_list) > 0, "No images ids found"

print(f"{len(all_images_id_list)} barcodes ids available in last {DAYS_BACK_RANGE} days.")
images = []
document_detector = detector.DocumentDetector()
root = '/data/doc_detect/barcodes/mlweb_dump/'
count = 0
total = 0
for image_idx, image_id in enumerate(all_images_id_list):
    try:
        image_url = urljoin(base_url, f'/images/{image_id}')
        # Send a GET request to fetch the image data
        response = session.get(image_url)
        image_bytes = io.BytesIO(response.content)
        image = PIL.Image.open(image_bytes)
        # Check if the image is in RGBA mode
        if image.mode == 'RGBA':
            image = image.convert('RGB')
        # images.append([image, image_id])
        result = document_detector.detect_document(name='test', image=image)
        if result[0] is True:
            pad_x = 20
            pad_y = 20
            box = result[1]['raw_result']['barcode']['original_box']
            im_arr = np.array(image)
            im_arr = im_arr[box[1]-pad_y:box[3]+pad_y, box[0]-pad_x:box[2]+pad_x]

            filename = image_id + '.jpg'
            if filename not in os.listdir(root):
                PIL.Image.fromarray(im_arr).save(root + filename)
                count += 1

            images.append([PIL.Image.fromarray(im_arr), image_id])
            total += 1
    except Exception as e:
        continue
        print('Error:', e)
print(count, total, 'new barcodes added')


from setuptools import setup, find_packages
import os

setup(
    name='barcode_reader',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version='1.3.7',
    packages=find_packages(),
    package_data={'barcode_reader.models': ['adder.pytorch.model']},
    include_package_data=True,
    install_requires=['torch>=1.13.1',
                      'torchvision',
                      'Pillow',
                      'zxing-cpp',
                      'piqa'
                    #   'autocapture_model'
                      ]
)
# cp /home/nathan/barcode-reader/dist/barcode_reader-1.0.25.tar.gz /var/lib/packages/barcode_reader-1.0.25.tar.gz
# python setup.py sdist
# pip install --extra-index-url https://pip.ml.prod.trulioo.com/nathan/dev --trusted-host https://pip.ml.prod.trulioo.com/ barcode_reader


import torch
from piqa import SSIM
import torchvision


class DecoderBlock(torch.nn.Module):
    def __init__(self, in_channels, inter, out_channels, dropout_rate, bias, last=False):
        super().__init__()
        self.last = last
        self.conv1 = torch.nn.Conv2d(in_channels=in_channels, out_channels=inter, kernel_size=3, padding=1, bias=bias)
        self.d1 = torch.nn.Dropout(p=dropout_rate)
        self.b1 = torch.nn.BatchNorm2d(inter)
        self.relu1 = torch.nn.LeakyReLU(0.1)
        self.conv2 = torch.nn.Conv2d(inter, inter, kernel_size=3, padding=1, bias=bias)
        self.d2 = torch.nn.Dropout(p=dropout_rate)
        self.b2 = torch.nn.BatchNorm2d(inter)
        self.relu2 = torch.nn.LeakyReLU(0.1)
        if not self.last:
            self.convT1 = torch.nn.ConvTranspose2d(inter, out_channels, kernel_size=2, stride=2)
        else:
            self.conv3 = torch.nn.Conv2d(in_channels=inter, out_channels=out_channels, kernel_size=1, bias=bias)

    def forward(self, x):
        x = self.conv1(x)
        x = self.b1(x)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.b2(x)
        x = self.relu2(x)
        x = self.d1(x)
        if not self.last:
            x = self.convT1(x)
        else:
            x = self.conv3(x)
        return x


class EncoderBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, dropout_rate, bias):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1, bias=bias)
        self.d1 = torch.nn.Dropout(p=dropout_rate)
        self.b1 = torch.nn.BatchNorm2d(out_channels)
        self.relu1 = torch.nn.LeakyReLU(0.1)
        self.conv2 = torch.nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, bias=bias)
        self.d2 = torch.nn.Dropout(p=dropout_rate)
        self.b2 = torch.nn.BatchNorm2d(out_channels)
        self.relu2 = torch.nn.LeakyReLU(0.1)

    def forward(self, x):
        x = self.conv1(x)
        x = self.d1(x)
        x = self.b1(x)
        x = self.relu1(x)
        x = self.conv2(x)
        x = self.d2(x)
        x = self.b2(x)
        x = self.relu2(x)
        return x


class UNet(torch.nn.Module):
    def __init__(self, in_channels=3, n_class=3, dropout_rate=0.0, magnitude=64):
        super().__init__()
        self.encoder_block1 = EncoderBlock(in_channels=in_channels, out_channels=magnitude, dropout_rate=dropout_rate, bias=True)
        self.encoder_block2 = EncoderBlock(in_channels=magnitude, out_channels=2*magnitude, dropout_rate=dropout_rate, bias=True)
        self.encoder_block3 = EncoderBlock(in_channels=2*magnitude, out_channels=4*magnitude, dropout_rate=dropout_rate, bias=True)
        self.encoder_block4 = EncoderBlock(in_channels=4*magnitude, out_channels=8*magnitude, dropout_rate=dropout_rate, bias=True)
        self.encoder_block5 = EncoderBlock(in_channels=8*magnitude, out_channels=16*magnitude, dropout_rate=dropout_rate, bias=True)
        self.bridge = torch.nn.ConvTranspose2d(16 * magnitude, 8 * magnitude, kernel_size=2, stride=2)
        self.pool = torch.nn.MaxPool2d(kernel_size=2, stride=2)
        self.decoder_block1 = DecoderBlock(in_channels=16*magnitude, inter=8*magnitude, out_channels=4*magnitude,
                                           dropout_rate=dropout_rate, bias=True)
        self.decoder_block2 = DecoderBlock(in_channels=8*magnitude, inter=4*magnitude, out_channels=2*magnitude,
                                           dropout_rate=dropout_rate, bias=True)
        self.decoder_block3 = DecoderBlock(in_channels=4*magnitude, inter=2*magnitude, out_channels=magnitude,
                                           dropout_rate=dropout_rate, bias=True)
        self.decoder_block4 = DecoderBlock(in_channels=2*magnitude, inter=magnitude, out_channels=n_class,
                                           dropout_rate=dropout_rate, bias=True, last=True)
        self.diff = torch.tensor(0.0)

    def forward(self, x):
        # Encoder
        residual = x
        e_block1 = self.encoder_block1(x)
        x = self.pool(e_block1)
        e_block2 = self.encoder_block2(x)
        x = self.pool(e_block2)
        e_block3 = self.encoder_block3(x)
        x = self.pool(e_block3)
        e_block4 = self.encoder_block4(x)
        x = self.pool(e_block4)
        x = self.encoder_block5(x)
        x = self.bridge(x)
        # Decoder
        x = torch.cat([x, e_block4], dim=1)
        x = self.decoder_block1(x)
        x = torch.cat([x, e_block3], dim=1)
        x = self.decoder_block2(x)
        x = torch.cat([x, e_block2], dim=1)
        x = self.decoder_block3(x)
        x = torch.cat([x, e_block1], dim=1)
        x = self.decoder_block4(x)
        self._regularize(x)
        x += residual
        return x

    def _regularize(self, x):
        self.diff = torch.norm(x)


class SSIMLoss(SSIM):
    def __init__(self):
        super().__init__()
        self.relu6 = torch.nn.ReLU6()

    def forward(self, x, y):
        return 1. - super().forward(self.relu6(x * 6.0) / 6.0, y)


class ReLU1(torch.nn.Module):
    def __init__(self, size):
        super().__init__()
        self.relu6 = torch.nn.ReLU6()
        self.size = size

    def forward(self, x):
        return self.relu6(x * 6.0) / 6.0


class L1_SSIM(torch.nn.Module):
    def __init__(self, alpha=0.5):
        super().__init__()
        self.alpha = alpha
        self.ssim_loss = SSIMLoss()
        self.l1_loss = torch.nn.L1Loss()

    def forward(self, x, target):
        ssim = self.ssim_loss(x, target)
        l1 = self.l1_loss(x, target)
        loss = self.alpha * ssim + (1 - self.alpha) * l1
        return loss.mean(), ssim.mean(), l1.mean()


class L1_DICE(torch.nn.Module):
    def __init__(self, alpha=0.5):
        super().__init__()
        self.alpha = alpha
        self.l1_loss = torch.nn.L1Loss()

    def forward(self, inputs, targets, smooth=1.):
        l1 = self.l1_loss(inputs, targets)
        inputs = inputs.view(-1)
        targets = targets.view(-1)
        intersection = (inputs * targets).sum()
        union = torch.sum(inputs * inputs) + torch.sum(targets * targets)
        dice_loss = 1 - (2.*intersection + smooth)/(union + smooth)
        loss = (self.alpha * (l1) + (1 - self.alpha) * dice_loss)
        return loss.mean(), dice_loss.mean(), l1.mean()


class VGGPerceptualLoss(torch.nn.Module):
    def __init__(self, resize=True):
        super(VGGPerceptualLoss, self).__init__()
        self.device = torch.device('cuda')
        blocks = []
        blocks.append(torchvision.models.vgg16(pretrained=True).features[:4].eval().to(self.device))
        blocks.append(torchvision.models.vgg16(pretrained=True).features[4:9].eval().to(self.device))
        blocks.append(torchvision.models.vgg16(pretrained=True).features[9:16].eval().to(self.device))
        blocks.append(torchvision.models.vgg16(pretrained=True).features[16:23].eval().to(self.device))
        for bl in blocks:
            for p in bl.parameters():
                p.requires_grad = False
        self.blocks = torch.nn.ModuleList(blocks).to(self.device)
        self.transform = torch.nn.functional.interpolate
        self.resize = resize
        self.register_buffer("mean", torch.tensor([0.485, 0.456, 0.406]).view(1, 3, 1, 1).to(self.device))
        self.register_buffer("std", torch.tensor([0.229, 0.224, 0.225]).view(1, 3, 1, 1).to(self.device))

    def forward(self, input, target, feature_layers=[0, 1, 2, 3], style_layers=[]):
        if input.shape[1] != 3:
            input = input.repeat(1, 3, 1, 1)
            target = target.repeat(1, 3, 1, 1)
        input = (input-self.mean) / self.std
        target = (target-self.mean) / self.std
        if self.resize:
            input = self.transform(input, mode='bilinear', size=(224, 224), align_corners=False)
            target = self.transform(target, mode='bilinear', size=(224, 224), align_corners=False)
        loss = 0.0
        x = input
        y = target
        for i, block in enumerate(self.blocks):
            x = block(x)
            y = block(y)
            if i in feature_layers:
                loss += torch.nn.functional.l1_loss(x, y)
            if i in style_layers:
                act_x = x.reshape(x.shape[0], x.shape[1], -1)
                act_y = y.reshape(y.shape[0], y.shape[1], -1)
                gram_x = act_x @ act_x.permute(0, 2, 1)
                gram_y = act_y @ act_y.permute(0, 2, 1)
                loss += torch.nn.functional.l1_loss(gram_x, gram_y)
        return loss, 0, 0


class BatchNormConv2d(torch.nn.Module):
    def __init__(self, input_channels, output_channels, resample_mode):
        super(BatchNormConv2d, self).__init__()
        self.input_channels = input_channels
        self.output_channels = output_channels

        if resample_mode == "down":
            self.kernel_size = (5, 5)
            self.stride = (2, 2)
            self.padding = (2, 2)
            self.conv = torch.nn.Conv2d(input_channels, output_channels, kernel_size=self.kernel_size,
                                        stride=self.stride, padding=self.padding)
        elif resample_mode == "up":
            self.kernel_size = (5, 5)
            self.stride = (2, 2)
            self.padding = (2, 2)
            self.out_padding = (1, 1)
            self.conv = torch.nn.ConvTranspose2d(input_channels, output_channels, kernel_size=self.kernel_size,
                                                 stride=self.stride, padding=self.padding, output_padding=self.out_padding)
        else:
            self.kernel_size = (3, 3)
            self.stride = (1, 1)
            self.padding = (1, 1)
            self.conv = torch.nn.Conv2d(input_channels, output_channels, kernel_size=self.kernel_size,
                                        stride=self.stride, padding=self.padding)

        self.norm = torch.nn.BatchNorm2d(output_channels)
        self.activation = torch.nn.SiLU()

    def forward(self, inputs):
        conv_out = self.conv(inputs)
        norm_out = self.norm(conv_out)
        return self.activation(norm_out)


class InvertedResidual(torch.nn.Module):
    def __init__(self, input_channels, expand_channels, squeeze_channels, downsample_flag):
        super(InvertedResidual, self).__init__()
        self.input_channels = input_channels
        self.expand_channels = expand_channels
        self.conv_expand = torch.nn.Conv2d(input_channels, expand_channels, kernel_size=(1, 1), stride=(1, 1))
        self.conv_contract = SeparableConv2d(expand_channels, input_channels, squeeze_channels, False)
        self.norm = torch.nn.BatchNorm2d(expand_channels)
        self.activation = torch.nn.SiLU()

    def forward(self, inputs):
        expand_out = self.conv_expand(inputs)
        expand_out = self.activation(self.norm(expand_out))
        return inputs + self.conv_contract(expand_out)


class SeparableConv2d(torch.nn.Module):
    def __init__(self, input_channels, output_channels, squeeze_channels, downsample_flag):
        super(SeparableConv2d, self).__init__()
        self.input_channels = input_channels
        self.conv_depthwise = torch.nn.Conv2d(input_channels, input_channels, kernel_size=(3, 3),
                                              stride=(1, 1), padding=(1, 1), groups=input_channels)
        self.norm_depthwise = torch.nn.BatchNorm2d(input_channels)
        self.activation = torch.nn.SiLU()
        self.squeeze_excite_flag = squeeze_channels > 0
        if self.squeeze_excite_flag:
            self.conv_reduce = torch.nn.Conv2d(input_channels, squeeze_channels, kernel_size=(1, 1), stride=(1, 1))
            self.conv_expand = torch.nn.Conv2d(squeeze_channels, input_channels, kernel_size=(1, 1), stride=(1, 1))
        if downsample_flag:
            self.conv_pointwise = torch.nn.Conv2d(input_channels, output_channels, kernel_size=(1, 1), stride=(2, 2))
        else:
            self.conv_pointwise = torch.nn.Conv2d(input_channels, output_channels, kernel_size=(1, 1), stride=(1, 1))
        self.norm_pointwise = torch.nn.BatchNorm2d(output_channels)
        self.downsample_flag = downsample_flag

    def forward(self, inputs):
        depthwise_out = self.conv_depthwise(inputs)
        depthwise_out = self.norm_depthwise(depthwise_out)
        depthwise_out = self.activation(depthwise_out)
        if self.squeeze_excite_flag is True:
            squeezed_out = torch.nn.functional.adaptive_avg_pool2d(depthwise_out, 1)
            squeezed_out = self.conv_reduce(squeezed_out)
            squeezed_out = self.activation(squeezed_out)
            squeezed_out = self.conv_expand(squeezed_out)
            squeezed_out = torch.sigmoid(squeezed_out)
            depthwise_out = squeezed_out * depthwise_out

        pointwise_out = self.conv_pointwise(depthwise_out)
        pointwise_out = self.norm_pointwise(pointwise_out)
        pointwise_out = self.activation(pointwise_out)
        return pointwise_out


class BarcodeDecoder(torch.nn.Module):
    def __init__(self):
        super(BarcodeDecoder, self).__init__()
        self.input_channels = 1
        self.activation = torch.nn.SiLU()
        self.init_num_channels = 64
        self.layers = torch.nn.Sequential()
        # self.layers.append(torch.nn.Linear(25600,51200))
        current_num_channels = self.init_num_channels * 2
        self.layers.append(torch.nn.Unflatten(1, (current_num_channels, 40, 160)))
        self.layers.append(InvertedResidual(current_num_channels, current_num_channels*2, 0, False))
        self.layers.append(BatchNormConv2d(current_num_channels, self.init_num_channels, "up"))
        self.layers.append(InvertedResidual(self.init_num_channels, self.init_num_channels*2, 0, False))
        self.layers.append(torch.nn.ConvTranspose2d(self.init_num_channels, self.input_channels,
                                                    kernel_size=(5, 5), stride=(2, 2), padding=(2, 2), output_padding=(1, 1)))
        self.layers.append(torch.nn.Tanh())

    def forward(self, inputs):
        return self.layers(inputs)


class BarcodeEncoder(torch.nn.Module):
    def __init__(self):
        super(BarcodeEncoder, self).__init__()
        self.input_channels = 1
        self.activation = torch.nn.SiLU()
        self.init_num_channels = 64
        self.layers = torch.nn.Sequential()
        current_num_channels = self.init_num_channels
        self.layers.append(BatchNormConv2d(self.input_channels, current_num_channels, "down"))
        self.layers.append(InvertedResidual(current_num_channels, current_num_channels*2, 0, False))
        self.layers.append(BatchNormConv2d(current_num_channels, current_num_channels*2, "down"))
        current_num_channels *= 2
        self.layers.append(InvertedResidual(current_num_channels, current_num_channels*2, 0, False))
        self.flatten = torch.nn.Flatten(start_dim=1)
        self.layers.append(self.flatten)
        # self.layers.append(torch.nn.Linear(51200,25600))

    def forward(self, inputs):
        return self.layers(inputs)
